import contextlib
import math
from abc import abstractmethod
from pathlib import Path

import yaml

from framjules import JulESConfig
from framjules.solve_handler.dataclasses import ComponentInfo, GraphInfos
from framjules.solve_handler.JulESNames import JulESNames


class ConfigHandler:
    """Responsible for implementing shared functionality in build method."""

    def __init__(
        self,
        folder: Path,
        config: JulESConfig,
        names: JulESNames,
        graph_infos: GraphInfos,
    ) -> None:
        """Responsible for building config file for JulES."""
        self.folder = folder
        self.config = config
        self.names = names
        self.graph_infos = graph_infos

    def configure(self) -> None:
        """Build JulES config and dump to YAML file."""
        config = self.get_default_config_dict()
        self._make_shrinkable(config)
        self.adapt_config_to_simulation_mode(config)
        self.adapt_to_dataset(config)  # TODO: Remove if JulES does not require this after all
        self.config_assert_no_missings(config)
        self.write_config_file(config)

    def is_clearing_only(self) -> bool:
        """Return True if dataset has no subsystems."""
        return all(not info.is_sss_member for info in self.graph_infos.clearing.values())

    def is_short_term_only(self) -> bool:
        """Return True if dataset has only short term storage subsystems."""
        has_sss = any(info.is_sss_member for info in self.graph_infos.clearing.values())
        no_long_term = all(info.is_sss_member and info.sss_is_short_term for info in self.graph_infos.clearing.values())
        return has_sss and no_long_term

    def adapt_to_dataset(self, config: dict) -> None:
        """Remove keys in config if dataset does require it."""
        n = self.names

        settings = config[n.PY_JULES_SETTINGS_NAME]

        if self.is_clearing_only():
            del settings[n.SCENARIO_GENERATION]
            del settings[n.SUBSYSTEMS]
            for sub_key in list(settings[n.PROBLEMS].keys()):
                if sub_key != n.CLEARING:
                    del settings[n.PROBLEMS][sub_key]
            del settings[n.RESULTS][n.SCENARIOS]
            del settings[n.RESULTS][n.STORAGEVALUES]
            settings[n.HORIZONS][n.COMMODITIES] = [n.MARKET]
            with contextlib.suppress(Exception):
                for sub_key in list(settings[n.HORIZONS].keys()):
                    if sub_key not in [n.CLEARING, n.COMMODITIES]:
                        del settings[n.HORIZONS][sub_key]
            with contextlib.suppress(Exception):
                del settings[n.HORIZONS][n.CLEARING][n.STORAGE_SYSTEM]
            del settings[n.TIME][n.PROB_TIME][n.PHASE_IN_TIME]
            del settings[n.TIME][n.PROB_TIME][n.PHASE_IN_DELTA_STEPS]
            del settings[n.TIME][n.PROB_TIME][n.PHASE_IN_DELTA_DAYS]

        elif self.is_short_term_only():
            del settings[n.SCENARIO_GENERATION]
            settings[n.HORIZONS][n.COMMODITIES] = [n.MARKET]
            for term in [n.CLEARING, n.SHORT, n.MED, n.LONG]:
                del settings[n.HORIZONS][term][n.STORAGE_SYSTEM]

        weather_year_start, num_scen_years = self.config.get_weather_years()
        weather_year_stop = weather_year_start + num_scen_years  # also check prognosis and evp scenarios
        if (
            n.SCENARIO_GENERATION in settings
            and settings[n.SCENARIO_GENERATION][n.STOCHASTIC][n.NUM_SCEN] >= weather_year_stop - weather_year_start
        ):
            del settings[n.SCENARIO_GENERATION]

    def _is_other_flow_market(self, x: ComponentInfo) -> bool:
        return x.is_flow and (not x.is_exogenous) and x.is_market_flow

    def _is_other_flow_hydro(self, x: ComponentInfo) -> bool:
        return x.is_flow and (not x.is_exogenous) and (not x.is_market_flow)

    def get_default_config_dict(self) -> dict:
        """Get default config."""
        n = self.names

        weather_year_start, num_scen_years = self.config.get_weather_years()
        weather_year_stop = weather_year_start + num_scen_years

        other_flows_market = [k for k, v in self.graph_infos.clearing.items() if self._is_other_flow_market(v)]
        other_flows_hydro = [k for k, v in self.graph_infos.clearing.items() if self._is_other_flow_hydro(v)]

        return {
            n.MAIN: {
                n.INPUT: str(self.folder),
                n.OUTPUT_PATH: str(self.folder),
                n.NUM_CORES: self.config.get_num_cpu_cores(),
                n.DATA_YEARS: n.MISSING_CONFIG,
                n.SCENARIO_YEARS: n.MISSING_CONFIG,
                n.WEEK_START: n.MISSING_CONFIG,
                n.NUM_SIM_YEARS: n.MISSING_CONFIG,
                n.SETTINGS: n.PY_JULES_SETTINGS_NAME,
                n.OUTPUT_INDEX: self.config_get_output_index(),
                n.DATETIME_FORMAT: n.DATETIME_FORMAT_JULESIO,
                n.OUTPUT_FORMAT: n.HDF5,
                n.EXTRA_STEPS: 0,
                n.ELASTIC: n.FALSE,
            },
            n.PY_JULES_SETTINGS_NAME: {
                n.TIME: {
                    n.WEATHER_YEAR_START: weather_year_start,
                    n.WEATHER_YEAR_STOP: weather_year_stop,
                    n.PROB_TIME: n.MISSING_CONFIG,
                    n.SKIPMAX: self._get_skipmax(),
                },
                n.PROBLEMS: {
                    n.PROGNOSIS: {
                        n.AGGSUPPLYN: 4,
                        n.RESIDUAL_AREA_LIST: [],
                        n.SHORT_TERM_STORAGE_CUTOFF_HOURS: self.config.get_short_term_storage_cutoff_hours(),
                        n.LONG: {
                            n.SOLVER: n.JUMP_HIGHS if self.config.get_debug_long_opt_solver() else n.HIGHS_SIMPLEX,
                        },
                        n.MED: {n.SOLVER: n.JUMP_HIGHS if self.config.get_debug_med_opt_solver() else n.HIGHS_SIMPLEX},
                        n.SHORT: {
                            n.SOLVER: n.JUMP_HIGHS
                            if self.config.get_debug_short_opt_solver()
                            else n.HIGHS_SIMPLEX_NO_WARMSTART,
                        },
                        n.STARTSTORAGES: {n.FUNCTION: n.AGG_STARTMAG_DICT},
                    },
                    n.ENDVALUE: {
                        n.SOLVER: n.JUMP_HIGHS if self.config.get_debug_end_value_opt_solver() else n.HIGHS_SIMPLEX,
                        n.STARTSTORAGES: {n.FUNCTION: n.STARTMAG_DICT},
                    },
                    n.STOCHASTIC: {
                        n.MAXCUTS: 13,
                        n.LB: -1000000000000.0,
                        n.RELTOL: 0.0001,
                        n.MASTER: {
                            n.SOLVER: n.JUMP_HIGHS
                            if self.config.get_debug_subsystem_master_opt_solver()
                            else n.HIGHS_SIMPLEX,
                        },
                        n.SUBS: {
                            n.SOLVER: n.JUMP_HIGHS
                            if self.config.get_debug_subsystem_sub_opt_solver()
                            else n.HIGHS_SIMPLEX,
                        },
                        n.STARTSTORAGES: {n.FUNCTION: n.STARTMAG_DICT},
                        n.DISTRIBUTION_METHOD_MP: n.BYSIZE,
                        n.DISTRIBUTION_METHOD_SP: n.WITHMP,
                    },
                    n.CLEARING: {
                        n.SOLVER: n.JUMP_HIGHS
                        if self.config.get_debug_clearing_opt_solver()
                        else n.HIGHS_SIMPLEX_SIP_NO_WARMSTART,
                        n.STATEDEPENDENT_PROD: n.FALSE,
                        n.STATEDEPENDENT_PUMP: n.FALSE,
                        n.HEADLOSSCOST: n.FALSE,
                    },
                },
                n.SCENARIO_GENERATION: {
                    n.STOCHASTIC: {
                        n.FUNCTION: n.INFLOW_CLUSTERING_METHOD,
                        n.NUM_SCEN: 7,
                        n.SCEN_DELTA_DAYS: 365,
                        n.PARTS: 4,
                    },
                },
                n.SUBSYSTEMS: {
                    n.FUNCTION: n.TWO_STORAGE_DURATION,
                    n.SHORT_TERM_STORAGE_CUTOFF_HOURS: self.config.get_short_term_storage_cutoff_hours(),
                    n.SHORT_STOCH_DURATION_HOURS: self._get_short_stoch_hours(),
                    n.LONG_STOCH_DURATION_DAYS: self._get_long_stoch_days(),
                    n.LONG_EV_DURATION_DAYS: self._get_long_ev_days(),
                },
                n.HORIZONS: {
                    n.COMMODITIES: [n.POWER, n.HYDRO],
                    n.LONG: {
                        n.TERM_DURATION_DAYS: self.config.get_time_resolution().get_long_days(),
                        n.HYDRO: {
                            n.FUNCTION: n.SEQUENTIAL_HORIZON,
                            n.PERIOD_DURATION_DAYS: self.config.get_time_resolution().get_long_storage_days(),
                        },
                        n.POWER: (
                            self._get_adaptive_horizon(
                                clusters=self.config.get_time_resolution().get_long_adaptive_blocks(),
                                unit_duration_hours=self.config.get_time_resolution().get_long_adaptive_hours(),
                            )
                            if self._has_exogenous_market_flows("long_term")
                            else {
                                n.FUNCTION: n.SEQUENTIAL_HORIZON,
                                # ok as fallback because long storage period is multiple of med storage period
                                n.PERIOD_DURATION_DAYS: self.config.get_time_resolution().get_med_storage_days(),
                            }
                        ),
                    },
                    n.MED: {
                        n.TERM_DURATION_DAYS: self.config.get_time_resolution().get_med_days(),
                        n.HYDRO: {
                            n.FUNCTION: n.SEQUENTIAL_HORIZON,
                            n.PERIOD_DURATION_DAYS: self.config.get_time_resolution().get_med_storage_days(),
                        },
                        n.POWER: (
                            self._get_adaptive_horizon(
                                clusters=self.config.get_time_resolution().get_med_adaptive_blocks(),
                                unit_duration_hours=self.config.get_time_resolution().get_long_adaptive_hours(),
                            )
                            if self._has_exogenous_market_flows("medium_term")
                            else {
                                n.FUNCTION: n.SEQUENTIAL_HORIZON,
                                n.PERIOD_DURATION_DAYS: 1,  # ok as fallback because we know we always have whole days
                            }
                        ),
                    },
                    n.SHORT: {
                        n.TERM_DURATION_HOURS: self._get_short_hours(),
                        n.HYDRO: {
                            n.FUNCTION: n.SEQUENTIAL_HORIZON,
                            n.PERIOD_DURATION_HOURS: self._get_short_storage_hours(),
                        },
                        n.POWER: {
                            n.FUNCTION: n.SEQUENTIAL_HORIZON,
                            n.PERIOD_DURATION_HOURS: self._get_short_market_hours(),
                        },
                    },
                    n.CLEARING: {
                        n.TERM_DURATION_HOURS: self._get_clearing_hours(),
                        n.HYDRO: {
                            n.FUNCTION: n.SEQUENTIAL_HORIZON,
                            n.PERIOD_DURATION_HOURS: self._get_clearing_storage_hours(),
                        },
                        n.POWER: {
                            n.FUNCTION: n.SEQUENTIAL_HORIZON,
                            n.PERIOD_DURATION_HOURS: self._get_clearing_market_hours(),
                        },
                    },
                },
                n.RESULTS: {
                    n.MAINRESULTS: n.ALL,
                    "hydrolevels_water": n.TRUE,
                    n.TIMES: n.TRUE,
                    n.SCENARIOS: n.FALSE,
                    n.MEMORY: n.FALSE,
                    n.STORAGEVALUES: n.TRUE,
                    n.STORAGEVALUES_ALL_PROBLEMS: n.FALSE,
                    n.OTHER_TERMS: {
                        n.VARS: {
                            n.MARKET: other_flows_market,
                            **({n.HYDRO: other_flows_hydro} if other_flows_hydro else {}),
                        },
                    },
                },
            },
        }

    def _make_shrinkable(self, config: dict) -> None:
        # TODO: make less hard coded
        two_model_years = 364 * 2
        if self.config.get_time_resolution().get_long_days() < two_model_years:
            return

        shrinkafter_days = self.config.get_time_resolution().get_clearing_days()
        shrinkafter_days += self.config.get_time_resolution().get_short_days()
        shrinkafter_days += self.config.get_time_resolution().get_med_days()

        shrinkatleast_days = 360

        n = self.names
        config[n.PY_JULES_SETTINGS_NAME][n.HORIZONS][n.SHRINKABLE] = {
            n.SHRINKAFTER_DAYS: shrinkafter_days,
            n.SHRINKATLEAST_DAYS: shrinkatleast_days,
        }
        config[n.PY_JULES_SETTINGS_NAME][n.HORIZONS][n.LONG][n.SHRINKABLE] = n.TRUE

    def _get_clearing_hours(self) -> int:
        return int(self.config.get_time_resolution().get_clearing_days() * 24)

    def _get_clearing_market_hours(self) -> int:
        return int(self.config.get_time_resolution().get_clearing_market_minutes() / 60)

    def _get_clearing_storage_hours(self) -> int:
        return int(self.config.get_time_resolution().get_clearing_storage_minutes() / 60)

    def _get_short_hours(self) -> int:
        return int(self.config.get_time_resolution().get_short_days() * 24)

    def _get_short_market_hours(self) -> int:
        return int(self.config.get_time_resolution().get_short_market_minutes() / 60)

    def _get_short_storage_hours(self) -> int:
        return int(self.config.get_time_resolution().get_short_storage_minutes() / 60)

    def _get_long_ev_days(self) -> int:
        days = self.config.get_time_resolution().get_clearing_days()
        days += self.config.get_time_resolution().get_short_days()
        days += self.config.get_time_resolution().get_med_days()
        days += self.config.get_time_resolution().get_ev_days()
        return days

    def _get_long_stoch_days(self) -> int:
        days = self.config.get_time_resolution().get_clearing_days()
        days += self.config.get_time_resolution().get_short_days()
        days += self.config.get_time_resolution().get_med_days()
        return days

    def _get_short_stoch_hours(self) -> int:
        hours = self._get_clearing_hours()
        hours += self._get_short_hours()
        return hours

    def write_config_file(self, config: dict) -> None:
        """Write config.yml into folder."""
        # TODO: Add comments behind each config name?
        with Path.open(self.folder / self.names.FILENAME_CONFIG, "w") as f:
            yaml.dump(config, f, indent=self.names.YAML_INDENT)

    def config_get_output_index(self) -> str:
        """Which time dimension to use in output files.

        If serial simulation, scenario time is used.
        If prognosis simulation, data time is used.
        """
        if self.config.is_simulation_mode_serial():
            return self.names.WEATHER_YEAR
        # TODO add below when prognosis is supported:
        # if self.config.is_simulation_mode_prognosis():
        #     return self.names.DATA_YEAR
        raise NotImplementedError

    def _get_adaptive_horizon(self, clusters: int, unit_duration_hours: int) -> dict:
        n = self.names
        return {
            n.FUNCTION: n.ADAPTIVE_HORIZON,
            n.MACRO: n.STORAGE_SYSTEM,
            n.RHSDATA: {
                # TODO: support user supplied variation
                # n.FUNCTION: n.DYNAMIC_EXOGEN_PRICE_AH_DATA,
                # n.BALANCE_RHSDATA: self.names.dummy_exogenous_balance_name,
                n.FUNCTION: n.DYNAMIC_RHS_AH_DATA,
                n.COMMODITY: n.MARKET,
            },
            n.RHSMETHOD: n.KMEANS_AH_METHOD,
            n.CLUSTERS: clusters,
            n.UNIT_DURATION_HOURS: unit_duration_hours,
        }

    def _has_exogenous_market_flows(self, graph: str) -> bool:
        assert graph in ["clearing", "short_term", "medium_term", "long_term"]
        graphs_info = getattr(self.graph_infos, graph)
        return any(x.is_flow and x.is_exogenous and x.is_market_flow for x in graphs_info.values())

    @abstractmethod
    def adapt_config_to_simulation_mode(self, config: dict) -> None:
        """Add simulation mode spesific modifications to config."""
        pass

    def config_assert_no_missings(self, value: object) -> None:
        """RuntimeError if config containes MISSING_CONFIG after adapt_config_to_simulation_mode."""
        errors = set()
        self._config_assert_no_missings(errors, value, "")
        if errors:
            raise RuntimeError("\n".join(errors))

    def _config_assert_no_missings(self, errors: set[str], value: object, key: str) -> None:
        if isinstance(value, dict):
            for k, v in value.items():
                self._config_assert_no_missings(errors, v, k)
        if value == self.names.MISSING_CONFIG:
            message = f"Required config {key} not set."
            errors.add(message)

    def _get_skipmax(self) -> int:
        clearing_days = self.config.get_time_resolution().get_clearing_days()
        skipmax_days = self.config.get_skipmax_days()
        if skipmax_days == 0:
            return 0
        return math.floor(skipmax_days / clearing_days)



"""Config object for JulES solver."""

from __future__ import annotations

from datetime import timedelta
from pathlib import Path

from framcore.aggregators import Aggregator, HydroAggregator
from framcore.solvers import SolverConfig

from framjules import JulESTimeResolution

_SUPPORTED_AGGREGATORS = [HydroAggregator]


class JulESConfig(SolverConfig):
    """Class containing all config for JulES."""

    def __init__(self) -> None:
        """Create new JulESConfig object."""
        super().__init__()

        self._julia_exe_path: Path | str | None = None
        self._julia_env_path: Path | str | None = None
        self._julia_depot_path: Path | str | None = None

        self._branch_jules: str | None = None
        self._branch_tulipa: str | None = None

        self._short_term_aggregations: list[Aggregator] = []

        self._time_resolution = JulESTimeResolution()

        self._duration_clearing = timedelta(days=1)
        self._duration_short_term = timedelta(days=6)
        self._duration_medium_term = timedelta(days=364)
        self._duration_long_term = timedelta(days=364 * 5)

        self._market_resolution_clearing = timedelta(hours=1)
        self._market_resolution_short_term = timedelta(hours=2)

        self._market_num_blocks_medium_term = 5
        self._market_num_blocks_long_term = 4

        self._storage_resolution_clearing = timedelta(days=1)
        self._storage_resolution_short_term = timedelta(days=2)
        self._storage_resolution_medium_term = timedelta(days=7)
        self._storage_resolution_long_term = timedelta(days=28)

        self._short_term_storage_cutoff_hours = 10

        self._debug_short_opt_solver = False
        self._debug_med_opt_solver = False
        self._debug_long_opt_solver = False
        self._debug_end_value_opt_solver = False
        self._debug_subsystem_master_opt_solver = False
        self._debug_subsystem_sub_opt_solver = False
        self._debug_clearing_opt_solver = False
        self._is_cache_db = True
        self._skip_install_dependencies = False

        self._clearing_days = 2
        self._market_duration_minutes = 6 * 60
        self._storage_duration_minutes = 2 * 24 * 60
        self._lookahead_days = 5 * 365
        self._detail_level = "fast"

        self._skipmax_days = 6
        self._warn_skipmax_days = 32

    def set_skipmax_days(self, days: int) -> None:
        """Set number of days between calculation of medium and long term storage values.

        This can speed up a simulation. The cost is less good storage values. The longer between
        re-calculation of storage values, the bigger negative impact on simulation result quality.

        If skipmax_days = 6 and clearing_days = 2, JulES will calculate medium and long
        term storage values every 3rd simulation step.
        """
        self._check_type(days, int)
        self._check_int(days, lower_bound=0, upper_bound=None)
        if days > self._warn_skipmax_days:
            message = (
                "Unusually high value for skipmax_days: "
                f"Medium and long term storage values updated only {days}th day. "
                "This can give poor simulation results due to poor storage utilization."
            )
            self.send_warning_event(message)
        self._skipmax_days = days

    def get_skipmax_days(self) -> int:
        """Get number of days between calculation of medium and long term storage values."""
        return self._skipmax_days

    def is_skip_install_dependencies(self) -> bool:
        """Return True if install dependencies will be skipped during by JulES.solve."""
        return self._skip_install_dependencies

    def activate_skip_install_dependencies(self) -> None:
        """Tell JulES to not install dependencies, assuming they are already installed.

        Default is to install.
        """
        self._skip_install_dependencies = True

    def deactivate_skip_install_dependencies(self) -> None:
        """Tell JulES to install dependencies. (This is the default)."""
        self._skip_install_dependencies = False

    def is_cache_db(self) -> bool:
        """Return True if JulES is allowed to use a cache to store precomputed values while building."""
        return self._is_cache_db

    def activate_cache_db(self) -> None:
        """Activates use of cache db."""
        self._is_cache_db = True

    def deactivate_cache_db(self) -> None:
        """Activates use of db without cache."""
        self._is_cache_db = False

    def get_time_resolution(self) -> JulESTimeResolution:
        """Get time resolution object. Modify this to modify time resolution of JulES."""
        return self._time_resolution

    def get_short_term_storage_cutoff_hours(self) -> int:
        """Return num hours.

        JulES will classify all storage subsystems with
        max storage duration less than cutoff as short term storage.
        """
        return self._short_term_storage_cutoff_hours

    def set_jules_version(self, jules_branch: str | None = None, tulipa_branch: str | None = None) -> None:
        """Set which git branch of JulES and/or TuLiPa to use."""
        self._check_type(jules_branch, (str, type(None)))
        self._check_type(tulipa_branch, (str, type(None)))
        if jules_branch is not None:
            self._branch_jules = jules_branch
        if tulipa_branch is not None:
            self._branch_tulipa = tulipa_branch
        if self._branch_tulipa is None and self._branch_jules is not None:
            self._branch_tulipa = self._branch_jules

    def get_jules_version(self) -> str | None:
        """Get JulES git branch."""
        return self._branch_jules

    def get_tulipa_version(self) -> str | None:
        """Get TuLiPa git branch."""
        return self._branch_tulipa

    def set_julia_depot_path(self, path: Path) -> None:
        """Set folder where Julia installs new packages."""
        self._check_type(path, Path)
        self._julia_depot_path = path

    def get_julia_depot_path(self) -> Path | None:
        """Get folder where Julia installs new packages."""
        return self._julia_depot_path

    def set_julia_env_path(self, path: Path) -> None:
        """Set which Julia environment to use."""
        self._check_type(path, Path)
        self._julia_env_path = path

    def get_julia_env_path(self) -> Path | None:
        """Get Julia environment being used."""
        return self._julia_env_path

    def set_julia_exe_path(self, path: Path) -> None:
        """Set which Julia installation to use."""
        self._check_type(path, Path)
        self._julia_exe_path = path

    def get_julia_exe_path(self) -> Path | None:
        """Get Julia installation being used."""
        return self._julia_exe_path

    def _check_supported_aggregators(self, aggregators: list[Aggregator]) -> None:
        for aggr in aggregators:
            if not isinstance(aggr, tuple(_SUPPORTED_AGGREGATORS)):
                message = f"Aggregator of type {type(aggr)} is not supported in JulES. Supported types are: {_SUPPORTED_AGGREGATORS}"
                raise TypeError(message)

    def set_short_term_aggregations(self, aggregators: list[Aggregator]) -> None:
        """Set aggregations to create the short term model from clearing (the Model object being solved)."""
        self._check_supported_aggregators(aggregators)
        self._short_term_aggregations = aggregators

    def get_short_term_aggregations(self) -> list[Aggregator]:
        """Get aggregations to create the short term model from clearing (the Model object being solved)."""
        return self._short_term_aggregations

    """
    Debug optimization solvers of the different JulES problems.
    Helpful if problems are encountered during solving, most commonly infeasibility issues.
    Replaces TuLiPa.HiGHS_Prob with TuLiPa.JuMP_Prob which has better debugging features:
    - More checks while building the optimization problem
    - If infeasible, solve the problem again with relaxed constraints (with penalties) and return the broken constraints
    - Outputs the optimization problem with variable and constraint names from FRAM
    """

    def set_debug_all_opt_solver(self, debug: bool) -> None:
        """Set whether to debug all optimization solvers."""
        self._debug_short_opt_solver = debug
        self._debug_med_opt_solver = debug
        self._debug_long_opt_solver = debug
        self._debug_end_value_opt_solver = debug
        self._debug_subsystem_master_opt_solver = debug
        self._debug_subsystem_sub_opt_solver = debug
        self._debug_clearing_opt_solver = debug

    def set_debug_short_opt_solver(self, debug: bool) -> None:
        """Set whether to debug the short-term optimization solver."""
        self._debug_short_opt_solver = debug

    def get_debug_short_opt_solver(self) -> bool:
        """Get whether to debug the short-term optimization solver."""
        return self._debug_short_opt_solver

    def set_debug_med_opt_solver(self, debug: bool) -> None:
        """Set whether to debug the medium-term optimization solver."""
        self._debug_med_opt_solver = debug

    def get_debug_med_opt_solver(self) -> bool:
        """Get whether to debug the medium-term optimization solver."""
        return self._debug_med_opt_solver

    def set_debug_long_opt_solver(self, debug: bool) -> None:
        """Set whether to debug the long-term optimization solver."""
        self._debug_long_opt_solver = debug

    def get_debug_long_opt_solver(self) -> bool:
        """Get whether to debug the long-term optimization solver."""
        return self._debug_long_opt_solver

    def set_debug_end_value_opt_solver(self, debug: bool) -> None:
        """Set whether to debug the end value optimization solver."""
        self._debug_end_value_opt_solver = debug

    def get_debug_end_value_opt_solver(self) -> bool:
        """Get whether to debug the end value optimization solver."""
        return self._debug_end_value_opt_solver

    def set_debug_subsystem_master_opt_solver(self, debug: bool) -> None:
        """Set whether to debug the subsystem master optimization solver."""
        self._debug_subsystem_master_opt_solver = debug

    def get_debug_subsystem_master_opt_solver(self) -> bool:
        """Get whether to debug the subsystem master optimization solver."""
        return self._debug_subsystem_master_opt_solver

    def set_debug_subsystem_sub_opt_solver(self, debug: bool) -> None:
        """Set whether to debug the subsystem sub optimization solver."""
        self._debug_subsystem_sub_opt_solver = debug

    def get_debug_subsystem_sub_opt_solver(self) -> bool:
        """Get whether to debug the subsystem sub optimization solver."""
        return self._debug_subsystem_sub_opt_solver

    def set_debug_clearing_opt_solver(self, debug: bool) -> None:
        """Set whether to debug the clearing optimization solver."""
        self._debug_clearing_opt_solver = debug

    def get_debug_clearing_opt_solver(self) -> bool:
        """Get whether to debug the clearing optimization solver."""
        return self._debug_clearing_opt_solver



"""JulES solver."""

import cProfile
from pathlib import Path
from time import time
import gc

from framcore import Model
from framcore.solvers import Solver
from framcore.loaders import Loader
from framcore.utils import add_loaders

from framjules import JulESConfig
from framjules.solve_handler.SolveHandler import SolveHandler

_PROFILE_PERFORMANCE = False


class JulES(Solver):
    """JulES solver."""

    def __init__(self) -> None:
        """Create new JulES solver with default config set."""
        super().__init__()
        self._config = JulESConfig()

    def get_config(self) -> JulESConfig:
        """Get internal config object. Modify this to configure JulES."""
        return self._config

    def _solve(
        self,
        folder: Path,
        model: Model,
    ) -> None:
        # TODO: verify that market resolution is finer than storage

        t0 = time()
        if _PROFILE_PERFORMANCE:
            profiler = cProfile.Profile()
            profiler.enable()
        handler = SolveHandler(folder, model, self.get_config())
        self.send_debug_event(f"SolveHandler time: {round(time() - t0, 2)} seconds")
        if _PROFILE_PERFORMANCE:
            profiler.disable()  # Stop profiling
            profiler.dump_stats("profile_solvehandler_init.prof")

        t = time()
        if _PROFILE_PERFORMANCE:
            profiler = cProfile.Profile()
            profiler.enable()
        handler.build()
        if _PROFILE_PERFORMANCE:
            profiler.disable()  # Stop profiling
            profiler.dump_stats("profile_solvehandler_build.prof")
        self.send_debug_event(f"build time: {round(time() - t, 2)} seconds")

        t = time()
        handler.configure()
        self.send_debug_event(f"configure time: {round(time() - t, 2)} seconds")

        t = time() 
        loaders: set[Loader] = set()
        add_loaders(loaders, model)
        for loader in loaders:
            loader.clear_cache()
        gc.collect()
        self.send_debug_event(f"clear_cache time: {round(time() - t, 2)} seconds")

        t = time() 
        handler.run()
        self.send_debug_event(f"run time: {round(time() - t, 2)} seconds")

        t = time()
        if _PROFILE_PERFORMANCE:
            profiler = cProfile.Profile()
            profiler.enable()
        handler.set_results()
        if _PROFILE_PERFORMANCE:
            profiler.disable()  # Stop profiling
            profiler.dump_stats("profile_solvehandler_results.prof")
        self.send_debug_event(f"set_results time: {round(time() - t, 2)} seconds")

        self.send_debug_event(f"JulES._solve time: {round(time() - t0, 2)} seconds")



# Rules which the given Aggregators must follow:
# Storages must be the same in all aggregations


from framcore import Base, Model
from framcore.aggregators import Aggregator
from framcore.components import Component, Flow, Node

# TODO: somehow check that storages are the same between short, medium and long term Models.
# Probably need to do this in its own function after the aggregation maps for simpler components are created.


class JulESAggregator(Base):
    def __init__(
        self,
        clearing: Model,
        short: list[Aggregator],
        medium: list[Aggregator],
        long: list[Aggregator],
    ) -> None:
        """Class for defining and calculation aggregated Model instances based on a clearing Model.

        Creates three aggregated Model instances for short-, medium- and long-term simulation in JulES.

        Note:
            - Short term Model is aggregated from a Clearing Model.
            - Medium term Model is aggregated from the Short term Model.
            - Long term Model is aggregated from the medium term Model

        Args:
            clearing (Model): The clearing Model to aggregate from.
            short (list[Aggregator]): List of aggregations to create the short term Model from clearing.
            medium (list[Aggregator]): List of aggregations to create the medium term Model from short
            long (list[Aggregator]): List of aggregations to create the long term Model from medium.

        """
        self._clearing = clearing
        self._short = short
        self._medium = medium
        self._long = long

        self._short_model: Model | None = None
        self._medium_model: Model | None = None
        self._long_model: Model | None = None

    def get_short_term_model(self) -> Model:
        """Apply defined aggregations for short term Model."""
        if self._short_model is None:
            self._short_model = self._aggregate(self._clearing, self._short)
        return self._short_model

    def get_medium_term_model(self) -> Model:
        """Apply defined aggregations for medium term Model."""
        if self._medium_model is None:
            self._medium_model = self._aggregate(self.get_short_term_model(), self._medium)
        return self._medium_model

    def get_long_term_model(self) -> Model:
        """Apply defined aggregations for long term Model."""
        if self._long_model is None:
            self._long_model = self._aggregate(self.get_medium_term_model(), self._long)
        return self._long_model

    def get_short_term_aggregation_map(self) -> dict[str, set[str] | None]:
        """Get the aggregation map of Components from clearing to short term Model."""
        return self._create_aggregation_map(self._clearing, self._short)

    def get_medium_term_aggregation_map(self) -> dict[str, set[str] | None]:
        """Get the aggregation map of Components from clearing to medium term Model."""
        return self._create_aggregation_map(self._clearing, self._short + self._medium)

    def get_long_term_aggregation_map(self) -> dict[str, set[str] | None]:
        """Get the aggregation map of Components from clearing to long term Model."""
        return self._create_aggregation_map(self._clearing, self._short + self._medium + self._long)

    def get_short_term_graph_map(
        self,
        graph_clearing: dict[str, Component],
        graph_short: dict[str, Component],
    ) -> dict[str, set[str] | None]:
        """Get aggregation map for version of short term Model with graph of Flows and Nodes."""
        return self._get_graph_aggregation_map(
            original_agg_map=self.get_short_term_aggregation_map(),
            clearing=self._clearing,
            graph_clearing=graph_clearing,
            aggregated=self.get_short_term_model(),
            graph_aggregated=graph_short,
        )

    def get_medium_term_graph_map(
        self,
        graph_clearing: dict[str, Component],
        graph_medium: dict[str, Component],
    ) -> dict[str, set[str] | None]:
        """Get aggregation map for version of medium term Model with graph of Flows and Nodes."""
        return self._get_graph_aggregation_map(
            original_agg_map=self.get_medium_term_aggregation_map(),
            clearing=self._clearing,
            graph_clearing=graph_clearing,
            aggregated=self.get_medium_term_model(),
            graph_aggregated=graph_medium,
        )

    def get_long_term_graph_map(
        self,
        graph_clearing: dict[str, Component],
        graph_long: dict[str, Component],
    ) -> dict[str, set[str] | None]:
        """Get aggregation map for version of long term Model with graph of Flows and Nodes."""
        return self._get_graph_aggregation_map(
            original_agg_map=self.get_long_term_aggregation_map(),
            clearing=self._clearing,
            graph_clearing=graph_clearing,
            aggregated=self.get_long_term_model(),
            graph_aggregated=graph_long,
        )

    def assert_equal_storages(
        self,
        simpler_short: dict[str, Component],
        simpler_medium: dict[str, Component],
        simpler_long: dict[str, Component],
    ) -> None:
        """Check that all Nodes with Storages are preserved between short, medium and long term Models.

        Args:
            simpler_short (dict[str, Component]): Short term Model Components.
            simpler_medium (dict[str, Component]): Medium term Model Components.
            simpler_long (dict[str, Component]): Long term Model Components.

        Raises:
            ValueError: If the Models have differing Storages.

        """
        short_storages = self._get_storages(simpler_short)
        medium_storages = self._get_storages(simpler_medium)
        long_storages = self._get_storages(simpler_long)

        if short_storages != medium_storages != long_storages:
            message = "Storages are not equal between short, medium and long term Models."
            unique_short = short_storages - (medium_storages | long_storages)
            unique_medium = medium_storages - (short_storages | long_storages)
            unique_long = long_storages - (short_storages | medium_storages)
            if unique_short:
                message += f"\n - Unique Nodes with Storages in Short Model: {unique_short}"
            if unique_medium:
                message += f"\n - Unique Nodes with Storages in Medium Model: {unique_medium}"
            if unique_long:
                message += f"\n - Unique Nodes with Storages in Long Model: {unique_long}"
            raise ValueError(message)

    def _aggregate(self, model: Model, aggs: list[Aggregator]) -> Model:
        if aggs:
            # works because aggregators should not modify the original components
            # except if disaggregate is called, but we shall only use aggregate
            agg_model = Model()
            agg_model.get_data().update(model.get_data())
        else:
            agg_model = model
        for agg in aggs:
            agg.aggregate(agg_model)
        return agg_model

    def _create_aggregation_map(self, clearing: Model, aggs: list[Aggregator]) -> dict[str, list[str]]:
        """Merge aggregation maps of a list of aggregators from clearing to final."""
        clearing_ids = [name for name, ob in clearing.get_data().items() if isinstance(ob, Component)]
        full_agg_mapping = {k: {v} for k, v in zip(clearing_ids, clearing_ids, strict=True)}

        for agg in aggs:
            agg_map = agg.get_aggregation_map()
            for detailed_id, aggregated_ids in full_agg_mapping.items():
                if not aggregated_ids:  # Component has been deleted by an aggregation.
                    continue

                new_agg_ids = set()
                for agg_id in aggregated_ids:
                    if agg_id not in agg_map:
                        new_agg_ids.add(agg_id)  # left as is
                        continue
                    if not agg_map[agg_id]:
                        # deleted. if all agg_ids are marked deleted, so is the detailed one.
                        continue
                    new_agg_ids |= agg_map[agg_id]

                full_agg_mapping[detailed_id] = new_agg_ids  # empty set signifies deleted component

        return full_agg_mapping

    def _get_graph_aggregation_map(
        self,
        original_agg_map: dict[str, set[str]],
        clearing: Model | dict[str, Component],
        graph_clearing: dict[str, Component],
        aggregated: Model | dict[str, Component],
        graph_aggregated: dict[str, Component],
    ) -> dict[str, set[str]]:
        """Create aggregation map with simpler Component IDs based on an original mapping from clearing to aggregated.

        Use get_top_parent of components to find IDs in original_agg_map then change to the Flow/Node ID.

        Args:
            original_agg_map (dict[str, set[str]]): Mapping between Components of clearing and aggregated Models.
            clearing (dict[str, Component]): Clearing Model with top parents.
            graph_clearing (dict[str, Component]): Clearing Model version with Flows and Nodes. Derived from
                                                     clearing.
            aggregated (dict[str, Component]): Aggregated Model with top parents. Aggregated from clearing.
            graph_aggregated (dict[str, Component]): Aggregated Model version with Flows and Nodes. Derived from
                                                       aggregated.

        Returns:
            dict[str, set[str]]: Mapping between components of simpler clearing and simpler aggregated Models.

        """
        if isinstance(clearing, Model):
            clearing = {k: v for k, v in clearing.get_data().items() if isinstance(v, Component)}
        if isinstance(aggregated, Model):
            aggregated = {k: v for k, v in aggregated.get_data().items() if isinstance(v, Component)}

        self._check_agg_map_compatibility(clearing, aggregated, original_agg_map)

        graph_clearing_map = self._get_top_parent_to_simple(original=clearing, simpler=graph_clearing)
        graph_aggregated_map = self._get_top_parent_to_simple(original=aggregated, simpler=graph_aggregated)
        simple_agg_map = {}

        for clearing_id, agg_ids in original_agg_map.items():
            # the two if statements are there for if we want to map only a subset of the simpler Components.
            if clearing_id in graph_clearing_map:
                if not agg_ids:
                    continue  # choose not to add deleted components. May change this later.
                simple_agg_ids = set()
                for agg_id in agg_ids:
                    if agg_id in graph_aggregated_map:  # Again to allow subset to be mapped
                        simple_agg_ids |= graph_aggregated_map[agg_id]  # add set if simple component ids
                if simple_agg_ids:
                    for graph_clearing_id in graph_clearing_map[clearing_id]:
                        simple_agg_map[graph_clearing_id] = simple_agg_ids

        self._check_agg_map_validity(graph_clearing, graph_aggregated, simple_agg_map)
        return simple_agg_map

    def _check_agg_map_compatibility(
        self,
        clearing: Model | dict[str, Component],
        aggregated: Model | dict[str, Component],
        original_agg_map: dict[str, set[str]],
    ) -> None:
        if set(clearing.keys()) != set(original_agg_map.keys()):
            missing_in_clearing = set(original_agg_map.keys()).difference(clearing.keys())
            extra_in_clearing = set(clearing.keys()).difference(original_agg_map.keys())
            message = (
                "clearing is incompatible with the aggregation mapping between clearing and aggregated Models.\n"
                f"Missing in clearing: {missing_in_clearing}\n"
                f"Extra in clearing: {extra_in_clearing}"
            )
            raise KeyError(message)

        original_agg_map_values = set().union(*(v for v in original_agg_map.values() if v))
        if set(aggregated.keys()) != original_agg_map_values:
            missing_in_aggregated = original_agg_map_values.difference(aggregated.keys())
            extra_in_aggregated = set(aggregated.keys()).difference(original_agg_map_values)
            message = (
                "aggregated is incompatible with the aggregation mapping between clearing and aggregated Models.\n"
                f"Missing in aggregated: {missing_in_aggregated}\n"
                f"Extra in aggregated: {extra_in_aggregated}"
            )
            raise KeyError(message)

    def _check_agg_map_validity(
        self,
        original_components: dict[str, Component],
        aggregated_components: dict[str, Component],
        agg_map: dict[str, set[str] | None],
    ) -> None:
        """Check Flow and Node rules for all mappings in an aggregation map."""
        errors = set()
        for original_id, aggregated_ids in agg_map.items():
            component = original_components[original_id]
            if isinstance(component, Node):
                self._check_node_rules(original_id, component, aggregated_ids, aggregated_components, errors)

            if isinstance(component, Flow) and component.get_startupcost() is not None:
                self._check_flow_rules(original_id, aggregated_ids, aggregated_components, errors)

            if not isinstance(component, (Flow, Node)):
                message = (
                    f"Invalig Model of simpler Components. Must consist of only Flows and Nodes. Found: {component}"
                )
                raise ValueError(message)

        self._report_errors(errors)

    def _check_node_rules(
        self,
        original_id: str,
        node: Node,
        aggregated_ids: set[str] | None,
        aggregated_components: dict[str, Component],
        errors: set[str],
    ) -> None:
        """Check rules for Nodes for a Component ID in an aggregation map.

        A Node on the disaggregated side (keys) must map to exactly one other Node. More keys are alowed to map to the
        same aggregated Node.

        """
        if node.get_storage() is None:
            # Check rules here?
            return

        if aggregated_ids is None:
            e = f"Node with Storage {original_id} was deleted during aggregations. This is not supported in JulES."
            errors.add(e)
            return
        aggregated_storages = set()
        for agg_id in aggregated_ids:
            agg_component = aggregated_components[agg_id]
            if isinstance(agg_component, Node) and agg_component.get_storage() is not None:
                aggregated_storages.add(agg_id)
        if len(aggregated_storages) != 1:
            errors.add(
                f"Node with Storage {original_id} must be connected to exactly one Node with Storage in the "
                f"aggregation map in JulES. Currently connected to: {aggregated_storages}.",
            )

    def _check_flow_rules(
        self,
        original_id: str,
        aggregated_ids: set[str] | None,
        aggregated_components: dict[str, Component],
        errors: set[str],
    ) -> None:
        """Check rules for Flows for a Component ID in an aggregation map.

        A Flow on the disaggregated side (keys) must map to exactly one other Flow. More keys are alowed to map to the
        same aggregated Flow.

        """
        if aggregated_ids is None:
            e = f"Flow with StartUpCost {original_id} was deleted during aggregations. This is not supported in JulES."
            errors.add(e)
            return
        aggregated_flows = set()
        for agg_id in aggregated_ids:
            agg_component = aggregated_components[agg_id]
            if isinstance(agg_component, Flow) and agg_component.get_startupcost() is not None:
                aggregated_flows.add(agg_id)
        if len(aggregated_flows) != 1:
            errors.add(
                f"Flow with StartUpCost {original_id} must be connected to exactly one Flow with StartUpCost in the "
                f"aggregation map in JulES. Currently connected to: {aggregated_flows}.",
            )

    @staticmethod
    def _get_storages(simpler: dict[str, Component]) -> set[str]:
        nodes_with_storages = set()
        for n, c in simpler.items():
            if isinstance(c, Node) and c.get_storage() is not None:
                nodes_with_storages.add(n)
        return nodes_with_storages

    @staticmethod
    def _get_top_parent_to_simple(original: dict[str, Component], simpler: dict[str, Component]) -> dict[str, set[str]]:
        """Map simpler components to their top parent."""
        inv_original = {c: n for n, c in original.items()}
        simpler_map: dict[str, set[str]] = {}

        for simple_id, component in simpler.items():
            top_parent = component.get_top_parent()
            if top_parent is None:
                message = (
                    f"Component {component} with ID {simple_id} has no parents. This means it has not been "
                    "derived from original."
                )
                raise ValueError(message)
            try:
                top_parent_id = inv_original[top_parent]
            except KeyError as e:
                message = (
                    f"Component {top_parent} does not exist in original Model. This means simpler has not been "
                    "derived from original."
                )
                raise KeyError(message) from e
            if top_parent_id in simpler_map:
                # list has been set, wo we add the simple component id to the
                simpler_map[top_parent_id].add(simple_id)
            else:
                simpler_map[top_parent_id] = {simple_id}

        return simpler_map
